/*
    linux loader
    http://www.linux-mips.org/wiki/Main_Page
*/
typedef unsigned char  UCHAR;
typedef unsigned short USHORT;
typedef unsigned long  ULONG;
#include <tap.h>
#include "exTap.h"
#include "tf5800.h"
#include "emma.h"

#define ID_USERTAP 0xbdb01504
TAP_ID( ID_USERTAP );

TAP_PROGRAM_NAME("linux_loader v0.04");
TAP_AUTHOR_NAME("bdb");
TAP_DESCRIPTION("a linux kernel loader");
TAP_ETCINFO(__DATE__);

//----------------------------------------------------------------------------
#define is_firmware(p) (((p&0xfc000000)==0x80000000) ? 1 : 0)

//----------------------------------------------------------------------------
//12.88
//dword fw_dcache_invalidate_1288 = 0x8000b864;
dword cmd_pbstop_1288 = 0x80239684;
dword DevDemux_TsSelDmxTs_1288 = 0x8007D398;
dword DevDemux_TsSelCamTs_1288 = 0x8007D524;

//13.26
//dword fw_dcache_invalidate_1326 = 0x8000D0D0;
dword cmd_pbstop_1326 = 0x80239684;
dword DevDemux_TsSelDmxTs_1326 = 0x8007EC04;
dword DevDemux_TsSelCamTs_1326 = 0x8007ED90;

//dword fw_dcache_invalidate;
dword cmd_pbstop;
dword DevDemux_TsSelDmxTs;
dword DevDemux_TsSelCamTs;

dword ramdisk_start;
dword ramdisk_end;
//----------------------------------------------------------------------------
//in load_elf.c
extern int load_elf(char *filename, dword *load, dword *entry, dword *length);
extern int load_root(char *filename, byte **rbuf, int *rlen);

//----------------------------------------------------------------------------
//
//tsr_commander stuff
//
#include "TSRCommander.inc"

extern void TSRCommanderConfigDialog();
extern bool TSRCommanderExitTAP();
extern int is_tsr_commander_running(void);

//----------------------------------------------------------------------------
//
//main
//
extern type_menu main_menu;
extern type_menu more_menu;

extern int TAP_Main(void);
extern dword TAP_EventHandler(word event, dword param1, dword param2);
extern int main_menu_handler(type_menu *menu, dword param1);
extern int more_menu_handler(type_menu *menu, dword param1);
extern int key_handler(dword param1);

extern int init(void);
extern int close(void);
extern int linux_bootloader(void);
extern int test1(void);
extern int test2(void);

byte *fbuf = 0;
int flen = 0;

//----------------------------------------------------------------------------
//
//tsr_commander stuff
//
bool TSRCommanderExitTAP()
{
    close();
    TAP_Exit();
    return FALSE;
}

void TSRCommanderConfigDialog()
{
    TAP_ExitNormal();
    ShowMenu(&main_menu);
}

int is_tsr_commander_running(void)
{
    int i;
    type_TAPProcess *TAPProcess;

    for (i=0; i<16; i++) {
        TAPProcess = (type_TAPProcess *)(fw_tap_table + (i*32));
        if (TAPProcess->firmwaregp) {
            if (TAPProcess->header) {
                if (TAPProcess->header->id == 0x807a10b2) {  //tsr_commander
                    return 1;
                }
                if (TAPProcess->header->id == 0x81001021) {  //tap_commander
                    return 1;
                }
            }
        }
    }
    return 0;
}

//----------------------------------------------------------------------------
//
//main
//
char *tap_name;
type_menu main_menu = {
    0,
    0,
    {0},
    0,
    (char *)__tap_program_name__,
    {
        {7, "xxx",              },
        {0, "Boot to Linux",    },
        {1, "test1",             },
        {2, "test2", },
        {3, "xxx", },
        {4, "xxx", },
        {5, "Exit Tap",         },
        {6, "More ...",         },
    },
};
type_menu more_menu = {
    0,
    0,
    {0},
    0,
    (char *)__tap_program_name__,
    {
        {8, "xxx",              },

        {0, "xxx",              },
        {1, "xxx",              },
        {2, "xxx",              },
        {3, "xxx",              },
        {4, "xxx",              },
        {5, "xxx",              },
        {6, "xxx",              },
        {7, "xxx",              },
    },
};

//-----------------------------------------------------------------------------
int TAP_Main(void)
{
    int ret;

    tap_name =  (char *)__tap_program_name__;
    sprintf(strLog, "Opening %s", tap_name);
    log_message(LOG_SCREEN_WAIT|LOG_SERIAL, strLog, DELAY_SHORT);

    //determine support for different firmwares
    ret = exTAP_Init();
    if (ret) {
        sprintf(strLog, "Unsupported platform - %d:%04x", fw_sysid, _appl_version);
        log_message(LOG_SCREEN_WAIT|LOG_SERIAL, strLog, DELAY_MEDIUM);
        return 0;
    }
    if ((fw_sysid==456)&&(_appl_version==0x1288)) {
//        fw_dcache_invalidate = fw_dcache_invalidate_1288;
        cmd_pbstop          = cmd_pbstop_1288;
        DevDemux_TsSelDmxTs = DevDemux_TsSelDmxTs_1288;
        DevDemux_TsSelCamTs = DevDemux_TsSelCamTs_1288;
    }
    else if ((fw_sysid==456)&&(_appl_version==0x1326)) {
//        fw_dcache_invalidate = fw_dcache_invalidate_1326;
        cmd_pbstop          = cmd_pbstop_1326;
        DevDemux_TsSelDmxTs = DevDemux_TsSelDmxTs_1326;
        DevDemux_TsSelCamTs = DevDemux_TsSelCamTs_1326;
    }
    else {
        sprintf(strLog, "Unsupported platform - %d:%04x", fw_sysid, _appl_version);
        log_message(LOG_SCREEN_WAIT|LOG_SERIAL, strLog, DELAY_MEDIUM);
        return 0;
    }

    init();
    return 1;                 //0=application mode, 1=tsr mode
}

dword TAP_EventHandler(word event, dword param1, dword param2)
{
    ULONG state, subState;
    int ret;

    switch (event) {
        case EVT_IDLE:
//            TAP_Print("\nEVT_IDLE, event=%x, param1=%x, param2=%x", event, param1, param2);
            if (message_showing) {
                if (TAP_GetTick() > message_showing) {
                    TAP_Osd_Delete(message_rgn);    // release rgn-handle
                    message_showing = 0;
                    message_rgn = 0;
                }
            }
            TSRCommanderWork();
            break;

        case EVT_KEY:
//            TAP_Print("\nEVT_KEY, event=%x, param1=%x, param2=%x", event, param1, param2);
            TAP_GetState(&state, &subState);
            if (main_menu.visible) { // handling keys when options window is shown
                param1 = main_menu_handler(&main_menu, param1);
            }
            else
            if (more_menu.visible) {
                param1 = more_menu_handler(&more_menu, param1);
            }
            else if ((state == STATE_Normal) && (subState == SUBSTATE_Normal)) {
                param1 = key_handler(param1);
            }
            break;
        case EVT_SVCSTATUS:
            break;
        case EVT_HOOK:
            break;
        default:
//            TAP_Print("\nEVT_???, event=%x, param1=%x, param2=%x", event, param1, param2);
            break;
    }
    return param1;
}

int main_menu_handler(type_menu *menu, dword param1)
{
    int selection;
    int ret;

    switch (param1) {
        case RKEY_Ok:
            selection = TAP_Win_GetSelection(&menu->win);
            DeleteMenu(menu);
            TAP_EnterNormal();
            switch (selection) {
                case 0:
                    ret = linux_bootloader();
                    if (ret) {
                        sprintf(strLog, "Linux BootLoader failed");
                        log_message(LOG_SCREEN_CONT|LOG_SERIAL, strLog, DELAY_SHORT);
                    }
                    return 0;
                case 1:
                    test1();
                    return 0;
                case 2:
                    test2();
//                    dump(0x83000000, 0x100);
                    return 0;
                case 3:
                    return 0;
                case 4:
                    return 0;
                case 5:
                    close();
                    TAP_Exit();
                    return 0;
                case 6:
                    TAP_ExitNormal();
                    ShowMenu(&more_menu);
                    return 0;
            }
            break;
        case RKEY_Recall:
            DeleteMenu(menu);
            TAP_EnterNormal();
            return param1; // cascades key to other TAP's
        case RKEY_Exit:
            DeleteMenu(menu);
            TAP_EnterNormal();
            return 0; // exit without Cascading
        default:
            TAP_Win_Action(&menu->win, param1); // send all other key's to menu-Win
            return 0;
    }
    return param1;
}

int more_menu_handler(type_menu *menu, dword param1)
{
    int selection;
    int num_events;
    int main_tuner, sub_tuner;
    int ret;
    byte ss_flag;

    switch (param1) {
        case RKEY_Ok:
            selection = TAP_Win_GetSelection(&menu->win);
            DeleteMenu(menu);
            TAP_EnterNormal();
            switch (selection) {
                case 0:
                    return 0;
                case 1:
                    return 0;
                case 2:
                    return 0;
                case 3:
                    return 0;
                case 4:
                    return 0;
                case 5:
                    return 0;
                case 6:
                    return 0;
                case 7:
                    return 0;
            }
            break;
        case RKEY_Recall:
            DeleteMenu(menu);
            TAP_EnterNormal();
            return param1; //RKEY_Recall; // cascades key to other TAP's
        case RKEY_Exit:
            DeleteMenu(menu);
            TAP_EnterNormal();
            return 0; // exit without Cascading
        default:
            TAP_Win_Action(&menu->win, param1); // send all other key's to menu-Win
            return 0;
    }
    return param1;
}

int key_handler(dword param1)
{
    static int seen_shutdown = 0;
    int chNum;

    switch(param1) {
        //only if tsrcommander is not running
        case RKEY_Recall:
            if (!is_tsr_commander_running()) {
                TAP_ExitNormal();
                ShowMenu(&main_menu);
                return 0;
            }
            break;
    }
    return param1;
}

int init(void)
{
    TAP_Print("\n_tap_startAddr=%x", _tap_startAddr);

    main_menu.visible = 0;
    more_menu.visible = 0;
    exTAP_Key_Extend2(1);

    TSRCommanderInit(0, TRUE);
    return 0;
}

int close(void)
{
    sprintf(strLog, "Closing  %s", tap_name);
    log_message(LOG_SERIAL|LOG_SCREEN_WAIT, strLog, DELAY_SHORT);

    if (fbuf) free(fbuf);

    exTAP_Key_Extend2(0);
    DeleteMenu(&more_menu);
    DeleteMenu(&main_menu);
    if (OpLog) TAP_Hdd_Fclose(OpLog);
    TSRCommanderErase();
    TAP_Print("\nClosing %s", tap_name);
    return 0;
}


//-----------------------------------------------------------------------------
static __inline void jmp_entry(dword entry, int argc, char **argv, char **envp, int *prom_vec)
{
    __asm__ __volatile__ (          //!!!beware the order!!!
    "or      $08,$00,%0\n"          //or      $t0,$00,kernel_entry
    "or      $16,$00,%1\n"          //or      $s0,$00,argc
    "or      $17,$00,%2\n"          //or      $s1,$00,argv
    "or      $18,$00,%3\n"          //or      $s2,$00,envp
    "or      $19,$00,%4\n"          //or      $s3,$00,prom_vec

    "move    $07,$19\n"             //move    $a3,$s3   prom_vec
    "move    $06,$18\n"             //move    $a2,$s2   envp
    "move    $05,$17\n"             //move    $a1,$s1   argv
    "move    $04,$16\n"             //move    $a0,$s0   argc
    "jr      $08\n"                 //j       $t0
    "nop\n"                         //!!!needed
    :
    : "r"(entry), "r"(argc), "r"(argv), "r"(envp), "r"(prom_vec)
    );
}

//does this work?
void dcache_invalidate()
{
    dword i;

//0 I Index_Invalidate
//    Set the cache state of the cache block to Invalid.
//0 D Index_Write_Back_Invalidate
//    Examine the cache state and W bit of the primary data cache block at the index
//    specified by the virtual address. If the state is not Invalid and the W bit is set, then
//    write back the block to memory. The address to write is taken from the primary cache
//    tag. Set cache state of primary cache block to Invalid.

    for (i=0x80000000; i<0x80002000; i+=0x10) {
        __asm__ __volatile__ (
            "or      $fp,$00,%0\n"          //or      $f0,$00,i
            //101111_reg_code_code_offset
            //101111_1e_0_1_0x0000
            //101111 $fp 1, 0000
//            ".byte 0xbf,0xc0,0x00,0x00\n"   //"cache 0, 0x0000($fp)"      //icache invalidate
            ".byte 0xbf,0xc1,0x00,0x00\n"   //"cache 1, 0x0000($fp)"      //dcache invalidate
        :: "r"(i));
    }
    //19=110 01
//6 D Hit_Write_Back
//    If the cache block contains the specified address, and the W bit is set, write back the
//    data to memory and clear the W bit.
}


int linux_bootloader(void)
{
    TYPE_PlayInfo pinfo;
    TYPE_RecInfo rinfo;
    int ret,i;
    byte *from;
    byte *to;
    dword kernel_load;
    dword kernel_entry;
    dword kernel_length;
    dword param[0x10];
    char *a[20];
    byte *rbuf;
    int rlen=0;
    char ramdisk_start_str[0x20];
    char ramdisk_end_str[0x20];

    TAP_Print("\nlinux bootloader()");

    //kill timeshift
    if (fw_stop_tshift>0) {
        TAP_Print("\nkilling timeshift");
//        TAP_SetSystemVar(SYSVAR_Timeshift, 0);
//        exTAP_CallFirmware((void *)fw_stop_tshift, 0,  0,0,0,0,0,0,0,0,0,0);
        exTAP_CallFirmware((void *)fw_stop_tshift, 1,  0,0,0,0,0,0,0,0,0,0);
    }

    //check if playing
    TAP_Hdd_GetPlayInfo(&pinfo);
    if (pinfo.playMode != PLAYMODE_None) {
        TAP_Print("\nerror: File is playing");
        return -1;
    }
    //check if recording
    TAP_Hdd_GetRecInfo(0, &rinfo);
    if (rinfo.recType != RECTYPE_None) {
        TAP_Print("\nerror: File is recording");
        return -1;
    }
    TAP_Hdd_GetRecInfo(1, &rinfo);
    if (rinfo.recType != RECTYPE_None) {
        TAP_Print("\nerror: File is recording");
        return -1;
    }


    //load the kernel into our memory space (fbuf, flen)
    TAP_Print("\nloading      : %s", "vmlinux");
    ret = load_elf("vmlinux", &kernel_load, &kernel_entry, &kernel_length);
    if (ret<0) {
        TAP_Print("\nerror: loading vmlinux");
        return ret;
    }

    //append rootfs.mips.ext2.gz
//    ret = load_root("rootfs.mips.cramfs", &rbuf, &rlen);
//    ret = load_root("rootfs.mips.ext2.gz", &rbuf, &rlen);
    ret = load_root("ramdisk.gz", &rbuf, &rlen);
    if (ret<0) {
        TAP_Print("\nerror: loading rootfs");
//        return ret;
    }

    TAP_Print("\nload address : %08x", kernel_load);
    TAP_Print("\ntotal_length : %8x",  kernel_length);
    TAP_Print("\nkernel_length: %8x",  flen);
    TAP_Print("\nrootfs_addr  : %08x", kernel_load+flen);
    TAP_Print("\nrootfs_length: %8x",  rlen);
    TAP_Print("\nkernel entry : %08x", kernel_entry);
    TAP_Print("\n");

    //shutdown everything we can
    //kill mheg
    if (fw_mheg_disable>0) {
        TAP_Print("\nkilling mheg");
//      exTAP_MHEG_Disable(1);
        exTAP_CallFirmware((void *)fw_mheg_disable, 0,0,0,0,0,0,0,0,0,0,0);
    }


    //kill epg
    TAP_Print("\nkilling epg");
    TAP_ControlEit(0);



    //kill channels
    TAP_Print("\nkilling channels");
    TAP_Channel_StopMainSub();


    //
    //this is a call to the 'pbstop' cmd which appears to shutdown a few ts mux bits
    //good for 12.88 only
    //seems to make booting much more reliable
    //
//    exTAP_CallFirmware((void *)cmd_pbstop, 0,0,0,0,0,0,0,0,0,0,0); -- crashes
    exTAP_CallFirmware((void *)DevDemux_TsSelDmxTs,0,0, 0,0,0,0,0,0,0,0,0);
    exTAP_CallFirmware((void *)DevDemux_TsSelDmxTs,1,2, 0,0,0,0,0,0,0,0,0);
    exTAP_CallFirmware((void *)DevDemux_TsSelDmxTs,2,1, 0,0,0,0,0,0,0,0,0);
    exTAP_CallFirmware((void *)DevDemux_TsSelCamTs,0,1, 0,0,0,0,0,0,0,0,0);


    //could wait for disk to spin down
    //TAP_Print("\n... wait for disk to spin down ...");
    TAP_Delay(100);


    //set the reboot flag to prevent verifying files+folders
//    TAP_Print("\nsetting clean shutdown flag");
//    *(word *)0xa3fffffc = 0x0000; //???
//    param[0] = 3;
//    if (fw_ioctl>0) exTAP_CallFirmware((void *)fw_ioctl, 0x02,(dword)&param,  0,0,0,0,0,0,0,0,0);     //update eeprom


    //kill remaining firmware
    //disable watchdog
//    TAP_Print("\nkilling watchdog");
//    if (fw_shutdown_irq_disable>0) exTAP_CallFirmware((void *)fw_shutdown_irq_disable, 0,0,0,0,0,0,0,0,0,0,0);
//    param[0] = 0x01;
//    if (fw_ioctl>0) exTAP_CallFirmware((void *)fw_ioctl, 0x00,0x07,(dword)&param,  0,0,0,0,0,0,0,0);


    //enter turbo mode - prevents a task switch
    TAP_Print("\nentering turbo mode");
    __asm__ __volatile__ (
    "sll     $26, 1"        //sll     $k0, 1
    );
    //li      $a0, 2
    //jal     sub_80000FC0


    //kill interrupts
    TAP_Print("\nkilling interrupts");
    *(dword *)EMMA_BHIF_NMI_EN   = 0x00;
    *(dword *)EMMA_BHIF_INT_EN_1 = 0x00;
    *(dword *)EMMA_BHIF_INT_EN_2 = 0x00;
//what does this do? - mask all intterupts @ processor
//    __asm__ __volatile__ (
//    "mfc0    $2, SR        \n"     //mfc0    $v0, SR
//    "ori     $4, $2, 0x400 \n"     //ori     $a0, $v0, 0x400   ; mask, im[0]
//    "mtc0    $4, SR        \n"     //mtc0    $a0, SR
//    );




//    if (fw_dcache_invalidate>0) exTAP_CallFirmware((void *)fw_dcache_invalidate, 0,0,0,0,0,0,0,0,0,0,0);


    //set the memory load address for the rootfs
    ramdisk_start = 0x83000000;
    ramdisk_end = ramdisk_start + rlen;
    sprintf(ramdisk_start_str, "ramdisk_start=0x%08x", ramdisk_start);
    sprintf(ramdisk_end_str, "ramdisk_end=0x%08x", ramdisk_end);


    TAP_Print("\nbooting linux ...\n\n");
   //copy the kernel into its runtime position - 0x80000000
    //no firmware code after here!!!!
    //memcpy(kernel_load, fbuf, flen);
    from = fbuf;
    to = (byte *)kernel_load;
    for (i=0; i<flen; i++) {
        *to++ = *from++;
    }
    //and append the rootfs
    from = rbuf;
    to = (byte *)ramdisk_start;
    for (i=0; i<rlen; i++) {
        *to++ = *from++;
    }


    //beware of the cache
    dcache_invalidate();


    //go for it
    //kernel command line arguments
    // The firmware/bootloader passes argc/argp/envp/prom_vec
    // argv should be 80000000-84000000
    //  vmlinux
    //  ramdisk_start=0x83000000   - set the start mem address of the rootfs
    //  ramdisk_end=0x83300000     - set the end mem address of the rootfs
    //  console=ttyS1,1152008N1    - enable the serial console
    //  mem=64M                    - enable 64M of ram
    //  root=/dev/ram0             - put rootfs on a ramdisk
    //  rw                         - make the ramdisk read/write
    //  init=/hello1               - pick a different starting point


//should make sure these parameters are in a 'nice' memory area
//
    a[0] = "vmlinux";
    a[1] = "ramdisk_start=0x83000000";        //now set earlier
    a[2] = "ramdisk_end=0x83400000";
//    a[1] = ramdisk_start_str;
//    a[2] = ramdisk_end_str;
    a[3] = "console=ttyS1,1152008N1";
    a[4] = "mem=64M";
//    a[4] = "mem=32";
    a[5] = "root=/dev/ram0";
    a[6] = "rw";
    a[7] = NULL;
    jmp_entry(kernel_entry, 7, a, NULL, NULL);

    TAP_Print("\n !!! should never get here !!!");
    return -1;
}



int test1()
{
    return 0;
}

int test2()
{
    dump(0xb0009000, 0x100);
    dump(0xb0009100, 0x100);
    dump(0xb000a000, 0x100);
    dump(0xb000a100, 0x100);
    dump(0xb000a200, 0x100);
    return 0;
}

void emma_init()
{
//is any of this worth doing again?
//maybe need to stop the dma?
//are the stream engines all stopped?
//audio dma?

//    *(dword *)EMMA_BHIF_STRAP_WRITE_2 = 0xfa;


    //ZPIO - Programmable Pin Interface
//    *(dword *)EMMA_ZPIO_PORT_WRITE_MASK_0 = 0;
//    *(dword *)EMMA_ZPIO_PORT_WRITE_MASK_1 = 0;

//    *(dword *)EMMA_ZPIO_PORT_DATA_0       = 0x0A002000;
//    *(dword *)EMMA_ZPIO_PORT_DATA_1       = 0;
//    *(dword *)EMMA_ZPIO_PORT_DIRECTION_0  = 0xFFC86A92;
//    *(dword *)EMMA_ZPIO_PORT_DIRECTION_1  = 0x7FF;

//    *(dword *)EMMA_ZPIO_PORT_DATA_0       = 0x0A002000;
//    *(dword *)EMMA_ZPIO_PORT_DATA_1       = 0;
//    *(dword *)EMMA_ZPIO_INT_MODE_0        = 0x20;
//    *(dword *)EMMA_ZPIO_INT_MODE_1        = 0;

//    *(dword *)EMMA_ZPIO_INT_CONDITION_A0  = 0;
//    *(dword *)EMMA_ZPIO_INT_CONDITION_A1  = 0x800;
//    *(dword *)EMMA_ZPIO_INT_CONDITION_B0  = 0x00100000;
//    *(dword *)EMMA_ZPIO_INT_CONDITION_B1  = 0;

    //PFUR - uart
//    baud = 1152000 / 115200;                   //1.152MHz xtal?
//    *(dword *)EMMA_PFUR1_IIR = 6;
//    *(dword *)EMMA_PFUR1_IIR = 1;
//    *(dword *)EMMA_PFUR1_LCR = 0x83;
//    *(dword *)EMMA_PFUR1_DIVH = (baud>>8) & 0xff;
//    *(dword *)EMMA_PFUR1_DIVL = (baud>>0) & 0xff;
//    *(dword *)EMMA_PFUR1_LCR = 3;
//    *(dword *)EMMA_PFUR1_IERD = 5;
//    *(dword *)EMMA_PFUR1_MCR = 0;


    //PURT - uart; front panel
//    *(dword *)EMMA_PURT_ERR   = 0;
//    *(dword *)EMMA_PURT_BRGMC = 0x24;
//    *(dword *)EMMA_PURT_BRGC  = 0xf0;
//    *(dword *)EMMA_PURT_ASIM  = 0xca;
//    *(dword *)EMMA_PURT_INT   = 0;
//    *(dword *)EMMA_PURT_INTM  = 0;
//    *(dword *)EMMA_PURT_DATA  = 0;


    //OGIO - GIO INTERFACE
//    *(dword *)EMMA_OGIO_AC_CNT = 0x3000034C;
//    *(dword *)0xB0000808 = 0;
//    *(dword *)0xB000080C = 0x0973;

//    *(dword *)EMMA_PCSI = 0;
//    *(dword *)(EMMA_PCIF+0x40) = 0;
//    *(dword *)(EMMA_PCIF+0x50) = 0;
//    *(dword *)EMMA_TTSD_PWM_DATA = 0x1FFF;

    //PIIC - i2c (tuner, video out)
//    *(dword *)(EMMA_PIIC0+0x20) = 3;
//    *(dword *)(EMMA_PIIC0+0x50) = 0;
//    *(dword *)(EMMA_PIIC0+0x60) = 0;
//    *(dword *)(EMMA_PIIC0+0x00) = 0x99;
//    *(dword *)(EMMA_PIIC1+0x20) = 3;
//    *(dword *)(EMMA_PIIC1+0x50) = 0;
//    *(dword *)(EMMA_PIIC1+0x60) = 0;
//    *(dword *)(EMMA_PIIC1+0x00) = 0x99;

    //ZTMR -
//    *(dword *)(EMMA_ZTMR0+00) = 2;
//    *(dword *)(EMMA_ZTMR0+00) = 0;
//    *(dword *)(EMMA_ZTMR0+14) = 0x28;
//    *(dword *)(EMMA_ZTMR0+04) = 0;
//    *(dword *)(EMMA_ZTMR0+18) = 0x23ff;
//    *(dword *)(EMMA_ZTMR0+00) = 1;

//    *(dword *)0xB000A004 = 1;
//    *(dword *)0xB000A008 = 1;

//    *(dword *)0xB0009000 = 0;
//    *(dword *)EMMA_XDMA_VBUF_RST = 0;
//    *(dword *)EMMA_XDMA_ABUF_RST = 0;
//    *(dword *)0xB000900C = 0;

//    *(dword *)EMMA_XDMA_INT_EN = 0;
//    *(dword *)EMMA_XDMA          = 0x0001;
//    *(dword *)EMMA_XDMA_VBUF_RST = 0x0101;
//    *(dword *)EMMA_XDMA_ABUF_RST = 0x0101;
//
//    *(dword *)EMMA_MPEG_RESET    = 1;
}
